/*

Brian Curless

Computer Graphics Laboratory
Stanford University

---------------------------------------------------------------------

Copyright (1997) The Board of Trustees of the Leland Stanford Junior
University. Except for commercial resale, lease, license or other
commercial transactions, permission is hereby given to use, copy,
modify this software for academic purposes only.  No part of this
software or any derivatives thereof may be used in the production of
computer models for resale or for use in a commercial
product. STANFORD MAKES NO REPRESENTATIONS OR WARRANTIES OF ANY KIND
CONCERNING THIS SOFTWARE.  No support is implied or provided.

*/

#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

#ifndef MAX
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif

#include <iostream>
#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/wait.h>

#include "vrip.h"
#include "vripGUICmds.h"
#include "vripGlobals.h"

#include <fstream.h>

// Use c++-ified version!!!
#include <tk.h>


static uchar *sliceBuf = NULL;
static int sliceXdim = 0, sliceYdim = 0;

static char *theRenderPhoto = NULL;
static uchar *renderBuf = NULL;
static int renderXdim = 0, renderYdim = 0;

static int thePhotoSliceArgc;
static char **thePhotoSliceArgv = NULL;
static Tcl_Interp *theInterp;

// Allocate fog, bog only once, to avoid memory leakage
// This has to be global, so that the multiple functions can all
// use the same front/back buffer.  Otherwise, changes from one
// filter will not be reflected in the others...  -lucas
static OccGrid *fog=NULL;
static OccGrid *bog=NULL;

int
Vrip_PhotoSliceCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{    
    char *axis;
    OccElement *slice;
    int sliceNum, xdim, ydim;

    if (argc != 4) {
	interp->result = "Usage: vrip_photoslice <photo-widget> <slice-num> <axis>";
	return TCL_ERROR;
    }

    if (theGrid == NULL) {
	interp->result = "Grid not allocated.";
	return TCL_ERROR;
    }
    
    Tk_PhotoHandle handle = Tk_FindPhoto(theInterp, argv[1]);
    if (handle == NULL)
	return TCL_ERROR;

    sliceNum = atoi(argv[2]);
    axis = argv[3];

    slice = theGrid->getSlice(axis, sliceNum, &xdim, &ydim);

    Tk_PhotoSetSize(handle, xdim, ydim);

    Tk_PhotoImageBlock block;
//    block.ptr = (uchar *)(slice + xdim*(ydim-1));
    block.pixelPtr = (uchar *)(slice);
    block.width = xdim;
    block.height = ydim;

    block.pitch = -xdim*4;
    block.pixelSize = 4;

#ifdef LINUX
    block.offset[0] = 1;
    block.offset[1] = 1;
    block.offset[2] = 1;
#else
    block.offset[0] = 0;
    block.offset[1] = 0;
    block.offset[2] = 0;
#endif

    Tk_PhotoPutBlock(handle, &block, 0, 0, xdim, ydim);

    return TCL_OK;

}


int
Vrip_SetPhotoSliceCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (thePhotoSliceArgv == NULL) {
	thePhotoSliceArgv = new char*[10];
	for (int i = 0; i < 10; i++) {
	    thePhotoSliceArgv[i] = new char[PATH_MAX];
	}
    }

    theInterp = interp;
    thePhotoSliceArgc = argc;
    for (int i = 0; i < argc; i++) {
	strcpy(thePhotoSliceArgv[i], argv[i]);
    }

    return TCL_OK;
}


void
updatePhotoSlice()
{
    Vrip_PhotoSliceRLECmd(NULL, theInterp, thePhotoSliceArgc, 
			 thePhotoSliceArgv);
}

int
Vrip_VolSizeCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
  sprintf(interp->result,"%d %d %d",frontRLEGrid->xdim,frontRLEGrid->ydim,frontRLEGrid->zdim);
  return TCL_OK;
}

/***************************************************************************
                   V R I P _ Q U A N T I Z E V O L C M D

Quantize the volume to either inside or outside. This will break the
smooth boundary and produce noticable voxel blocking in the output ply file
***************************************************************************/
int Vrip_QuantizeVolCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])

{
  if (frontRLEGrid == NULL) {
    interp->result = "Grid not allocated.";
    return TCL_ERROR;
  }

  int xmin=1,ymin=1,zmin=1;
  int xmax=frontRLEGrid->xdim-1;
  int ymax=frontRLEGrid->ydim-1;
  int zmax=frontRLEGrid->zdim-1;

  if (argc>6)
    {
      xmin=MAX(xmin,atoi(argv[1]));
      ymin=MAX(ymin,atoi(argv[2]));
      zmin=MAX(zmin,atoi(argv[3]));
      xmax=MIN(xmax,atoi(argv[4]));
      ymax=MIN(ymax,atoi(argv[5]));
      zmax=MIN(zmax,atoi(argv[6]));
    }
  
  printf ("Quantizing %d %d, %d %d, %d %d\n",xmin,xmax,ymin,ymax,zmin,zmax);

  //backRLEGrid->copy(frontRLEGrid);
  
  OccElement *line;

  static OccGrid *fog=frontRLEGrid->voxExpand();
  //static OccGrid *bog=frontRLEGrid->voxExpand();
  
  //  for (int y=1;y< frontRLEGrid->ydim-1;y++)
  for (int y=ymin;y<ymax;y++)
    {
      std::cout << y << " ";
      for (int z=zmin;z< zmax; z++)
	for (int x=xmin;x<xmax; x++)
	  {
	    OccElement el;
	    el=*(fog->address(x,y,z));
	    if(el.totalWeight>0)
	      {
		if (el.value<32767)
		  el.value=0;
		else
		  el.value=65535;
		
		*(fog->address(x,y,z))=el;
	      }
		 
	  }
    }

  frontRLEGrid->voxRLE(fog);


  return TCL_OK;
  
}
//////////////////////////////////////////////////////////////////////
// Feb 4, 1999
// Hole filling addition:
// James decides he wants to try blurring the volume to fill holes.
// This basically does one interation:
//    - Convert the volume to a non-RLE grid (front & back buffers)
//    - Reads from front (fog), writes to back (bog)
//    - blurs on the 3x3x3 cube, box filter
//    - writes it back out
// *note the following function, Lucas' version, which is based on
//  this one, but then he'll try making it more elaborate.
//////////////////////////////////////////////////////////////////////
int
Vrip_MuckWithVolCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
  if (frontRLEGrid == NULL) {
    interp->result = "Grid not allocated.";
    return TCL_ERROR;
  }

  int xmin=1,ymin=1,zmin=1;
  int xmax=frontRLEGrid->xdim-1;
  int ymax=frontRLEGrid->ydim-1;
  int zmax=frontRLEGrid->zdim-1;

  if (argc>6)
    {
      xmin=MAX(xmin,atoi(argv[1]));
      ymin=MAX(ymin,atoi(argv[2]));
      zmin=MAX(zmin,atoi(argv[3]));
      xmax=MIN(xmax,atoi(argv[4]));
      ymax=MIN(ymax,atoi(argv[5]));
      zmax=MIN(zmax,atoi(argv[6]));
    }
  
  printf ("Smoothing %d %d, %d %d, %d %d\n",xmin,xmax,ymin,ymax,zmin,zmax);


  //backRLEGrid->copy(frontRLEGrid);
  
  OccElement *line;



  // Allocate fog, bog only once, to avoid memory leakage
  // Otherwise, make bog identical to fog, in case we don't write every
  // pixel later...
  if (fog == NULL) { fog=frontRLEGrid->voxExpand(); }
  if (bog == NULL) { bog=frontRLEGrid->voxExpand(); }
  else { bog->copy(fog); }
		     

  
  //  for (int y=1;y< frontRLEGrid->ydim-1;y++)
  for (int y=ymin;y<ymax;y++)
    {
      std::cout << y << " ";
      for (int z=zmin;z< zmax; z++)
	for (int x=xmin;x<xmax; x++)
	  {
	    OccElement el;
	    el=*(fog->address(x,y,z));
	    if (el.totalWeight < 10)
	      {
		el.totalWeight=5;
		int totval=0;
		int totweight=0;
		//int ny=0;
		for (int nx= -1;nx<=1;nx++)
		  for (int ny= -1;ny<=1;ny++)
		    for (int nz =  -1;nz<=1;nz++)
		      {
			OccElement neighbor;
			neighbor = *(fog->address(x+nx,y+ny,z+nz));
			if (neighbor.totalWeight!=0)
			  {
			    totval+=neighbor.value;
			    totweight++;
			  }
		      }
		if (totweight>0)
		  {
		    el.value=(totval/totweight);
		    *(bog->address(x,y,z))=el;
		  }
	      }
	  }
    }

  frontRLEGrid->voxRLE(bog);

  // Don't free fog and bog, just swap 'em, so that we can use em again.
  // Since we wrote to bog, it's now the "good" one.  So swap fog and
  // bog, so next iteration, fog will be the up-to-date one....
  OccGrid *tog = fog;
  fog = bog;
  bog = tog;

  return TCL_OK;
  
}

//////////////////////////////////////////////////////////////////////
// Feb 4, 1999
// Hole filling addition:
// James decides he wants to try blurring the volume to fill holes.
// He initially wrote Vrip_MuckWithVolCmd.  Since it almost kinda
// worked, I didn't want to mess it up, so I made my own version of
// it.
// This basically does one interation:
//    - Convert the volume to a non-RLE grid (front & back buffers)
//    - Reads from front (fog), writes to back (bog)
//    - blurs on the 5x5x5 cube, box kernel, gaussian weighted filter
//    - finds dvaldx, dvaldy, dvaldz, to continue isolines?
//    - writes it back out
//////////////////////////////////////////////////////////////////////
int
Vrip_IsoGradBlurVolCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
  if (frontRLEGrid == NULL) {
    interp->result = "Grid not allocated.";
    return TCL_ERROR;
  }

  // Limits so that when we add/subtract 2, we don't walk off the volume
  int xmin=2,ymin=2,zmin=2;
  int xmax=frontRLEGrid->xdim-2;
  int ymax=frontRLEGrid->ydim-2;
  int zmax=frontRLEGrid->zdim-2;

  if (argc>6)
    {
      xmin=MAX(xmin,atoi(argv[1]));
      ymin=MAX(ymin,atoi(argv[2]));
      zmin=MAX(zmin,atoi(argv[3]));
      xmax=MIN(xmax,atoi(argv[4]));
      ymax=MIN(ymax,atoi(argv[5]));
      zmax=MIN(zmax,atoi(argv[6]));
    }
  
  printf ("Isogradient Smoothing %d %d, %d %d, %d %d\n",xmin,xmax,ymin,ymax,zmin,zmax);

  // Allocate fog, bog only once, to avoid memory leakage
  // Otherwise, make bog identical to fog, in case we don't write every
  // pixel later...
  if (fog == NULL) { fog=frontRLEGrid->voxExpand(); }
  if (bog == NULL) { bog=frontRLEGrid->voxExpand(); }
  else { bog->copy(fog); }

  static int voxCount = 0;

  for (int y=ymin;y<ymax;y++) {
    std::cout << y << " ";
    for (int z=zmin;z< zmax; z++) {
      for (int x=xmin;x<xmax; x++) {
	OccElement el;
	el=*(fog->address(x,y,z));

	/*
	// Randomly print out voxels...
	if ((voxCount++ % 79) == 0 && el.value == 0 &&
	    !(el.value == 65535 && el.totalWeight == 0)) {
	  cerr << "val: " << el.value << ", wt: " << el.totalWeight << endl;
	}
	*/

	// Only consider changing a voxel if it does not 
	// already have a high confidence...
	if (el.totalWeight < 10) {
	  // Set weight to 5 for now, it should really be 
	  //some nice falloff...
	  el.totalWeight=5;

	  // what we want to do here is compute the gradient on a 5x5 voxel grid.  
	  // for now, we'll weight every voxel, so that far-away voxels have less 
	  // effect than near voxels.
	  float gaussian[5] = {.13, .22, .3, .22, .13};

	  float totval = 0;
	  // Center of mass, moment of inertia:
	  // These two values are used to compute the gradients in x, y, z
	  // Used to compute center of mass of available data (com)
	  float comx=0, comy=0, comz=0;
	  // Center of mass squared
	  float com2x=0, com2y=0, com2z=0;

	  // Used to compute moment of inertia (moi)
	  float moix = 0, moiy = 0, moiz = 0;

	  float totweight = 0;
	  OccElement *neighbor;

	  // first pass: compute the average value, center of mass...
	  // (needed for second pass to find slope...)
	  int nx;
	  for (nx= -2;nx<=2;nx++) {
	    // x weight for this slice in x
	    float xweight = gaussian[nx+2];
	    for (int ny= -2;ny<=2;ny++) {
	      float xyweight = xweight * gaussian[ny+2];
	      for (int nz =  -2;nz<=2;nz++) {
		float xyzweight = xyweight * gaussian[nz+2];
		neighbor= fog->address(x+nx,y+ny,z+nz);
		if (neighbor->totalWeight != 0) {
		  float weightedVal = neighbor->value * xyzweight;
		  totval+= weightedVal;
		  totweight += xyzweight;
		  // Add this voxel to center of mass number
		  comx += xyzweight * nx;
		  comy += xyzweight * ny;
		  comz += xyzweight * nz;
		}
	      }
	    }
	  }

	  // Compute average val, center of mass (normalized)
	  float avgval = totval / totweight;
	  comx /= totweight;
	  comy /= totweight;
	  comz /= totweight;

	  // Second pass: compute the center-of-mass squared,
	  // and the moment of inertia of the value.  This requires
	  // already knowing the center of mass?
	  for (nx= -2;nx<=2;nx++) {
	    // x weight for this slice in x
	    float xweight = gaussian[nx+2];
	    for (int ny= -2;ny<=2;ny++) {
	      float xyweight = xweight * gaussian[ny+2];
	      for (int nz =  -2;nz<=2;nz++) {
		float xyzweight = xyweight * gaussian[nz+2];
		neighbor= fog->address(x+nx,y+ny,z+nz);
		if (neighbor->totalWeight != 0) {
		  float weightedVal = neighbor->value * xyzweight;

		  // Add this voxel to center of mass squared number
		  com2x += xyzweight * (nx - comx) * (nx - comx);
		  com2y += xyzweight * (ny - comy) * (ny - comy);
		  com2z += xyzweight * (nz - comz) * (nz - comz);
		  // Add this voxel to moment of inertia
		  moix += (nx-comx) * weightedVal;
		  moiy += (ny-comy) * weightedVal;
		  moiz += (nz-comz) * weightedVal;
		}
	      }
	    }
	  }
	  if (totweight>0.16) {
	    // Compute partial derivatives of value in each dir
	    float dvaldx = (com2x) ? (moix / com2x) : 0;
	    float dvaldy = (com2y) ? (moiy / com2y) : 0;
	    float dvaldz = (com2z) ? (moiz / com2z) : 0;
	    el.value = MAX(0, MIN(65535, avgval - 0.8 * ( dvaldx * comx +
						   dvaldy * comy +
						   dvaldz * comz)));
	    *(bog->address(x,y,z))=el;
	  }
	}
      }
    }
  }
  frontRLEGrid->voxRLE(bog);

  // Don't free fog and bog, just swap 'em, so that we can use em again.
  // Since we wrote to bog, it's now the "good" one.  So swap fog and
  // bog, so next iteration, fog will be the up-to-date one....
  OccGrid *tog = fog;
  fog = bog;
  bog = tog;

  return TCL_OK;
  
}

//////////////////////////////////////////////////////////////////////
// Feb 4, 1999
// Hole filling addition:
// James decides he wants to try blurring the volume to fill holes.
// He initially wrote Vrip_MuckWithVolCmd.  Since it almost kinda
// worked, I didn't want to mess it up, so I made my own version of
// it.
// This basically does one interation:
//    - Convert the volume to a non-RLE grid (front & back buffers)
//    - Reads from front (fog), writes to back (bog)
//    - blurs on the 3x3x3 cube, box filter
//    - writes it back out
//////////////////////////////////////////////////////////////////////
int
Vrip_GaussBlurVolCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
  if (frontRLEGrid == NULL) {
    interp->result = "Grid not allocated.";
    return TCL_ERROR;
  }

  // Limits so that when we add/subtract 2, we don't walk off the volume
  int xmin=2,ymin=2,zmin=2;
  int xmax=frontRLEGrid->xdim-2;
  int ymax=frontRLEGrid->ydim-2;
  int zmax=frontRLEGrid->zdim-2;

  if (argc>6)
    {
      xmin=MAX(xmin,atoi(argv[1]));
      ymin=MAX(ymin,atoi(argv[2]));
      zmin=MAX(zmin,atoi(argv[3]));
      xmax=MIN(xmax,atoi(argv[4]));
      ymax=MIN(ymax,atoi(argv[5]));
      zmax=MIN(zmax,atoi(argv[6]));
    }
  
  printf ("Gaussian Smoothing %d %d, %d %d, %d %d\n",xmin,xmax,ymin,ymax,zmin,zmax);

  // Allocate fog, bog only once, to avoid memory leakage
  // Otherwise, make bog identical to fog, in case we don't write every
  // pixel later...
  if (fog == NULL) { fog=frontRLEGrid->voxExpand(); }
  if (bog == NULL) { bog=frontRLEGrid->voxExpand(); }
  else { bog->copy(fog); }

  for (int y=ymin;y<ymax;y++) {
    std::cout << y << " ";
    for (int z=zmin;z< zmax; z++) {
      for (int x=xmin;x<xmax; x++) {
	OccElement el;
	el=*(fog->address(x,y,z));

	// Only consider changing a voxel if it does not 
	// already have a high confidence...
	if (el.totalWeight < 10) {
	  // Set weight to 5 for now, it should really be 
	  //some nice falloff...
	  el.totalWeight=5;

	  // what we want to do here is compute the gradient on a 5x5 voxel grid.  
	  // for now, we'll weight every voxel, so that far-away voxels have less 
	  // effect than near voxels.
	  float gaussian[5] = {.13, .22, .3, .22, .13};

	  float totval = 0;
	  float totweight = 0;
	  OccElement *neighbor;

	  // first pass: compute the average value, center of mass...
	  // (needed for second pass to find slope...)
	  for (int nx= -2;nx<=2;nx++) {
	    // x weight for this slice in x
	    float xweight = gaussian[nx+2];
	    for (int ny= -2;ny<=2;ny++) {
	      float xyweight = xweight * gaussian[ny+2];
	      for (int nz =  -2;nz<=2;nz++) {
		float xyzweight = xyweight * gaussian[nz+2];
		neighbor= fog->address(x+nx,y+ny,z+nz);
		if (neighbor->totalWeight != 0) {
		  float weightedVal = neighbor->value * xyzweight;
		  totval+= weightedVal;
		  totweight += xyzweight;
		}
	      }
	    }
	  }

	  float avgval = totval / totweight;

	  if (totweight> 0.1) {
	    el.value = MAX(0, MIN(65535, avgval));
	    *(bog->address(x,y,z))=el;
	  }
	}
      }
    }
  }
  frontRLEGrid->voxRLE(bog);

  // Don't free fog and bog, just swap 'em, so that we can use em again.
  // Since we wrote to bog, it's now the "good" one.  So swap fog and
  // bog, so next iteration, fog will be the up-to-date one....
  OccGrid *tog = fog;
  fog = bog;
  bog = tog;

  return TCL_OK;
  
}


//////////////////////////////////////////////////////////////////////
// Feb 9, 1999
// Hole filling addition:
//   Sharpen the volume
//   affect only the low-confidence voxels?
//   works directly on fog, so doesn't need to swapbuffers...

//////////////////////////////////////////////////////////////////////
int
Vrip_SharpenVolCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
  if (frontRLEGrid == NULL) {
    interp->result = "Grid not allocated.";
    return TCL_ERROR;
  }

  // Limits so that when we add/subtract 2, we don't walk off the volume
  int xmin=2,ymin=2,zmin=2;
  int xmax=frontRLEGrid->xdim-2;
  int ymax=frontRLEGrid->ydim-2;
  int zmax=frontRLEGrid->zdim-2;

  if (argc>6)
    {
      xmin=MAX(xmin,atoi(argv[1]));
      ymin=MAX(ymin,atoi(argv[2]));
      zmin=MAX(zmin,atoi(argv[3]));
      xmax=MIN(xmax,atoi(argv[4]));
      ymax=MIN(ymax,atoi(argv[5]));
      zmax=MIN(zmax,atoi(argv[6]));
    }
  
  printf ("Sharpening %d %d, %d %d, %d %d\n",xmin,xmax,ymin,ymax,zmin,zmax);

  // Allocate fog, bog only once, to avoid memory leakage
  if (fog == NULL) fog=frontRLEGrid->voxExpand(); 

  for (int y=ymin;y<ymax;y++) {
    std::cout << y << " ";
    for (int z=zmin;z< zmax; z++) {
      for (int x=xmin;x<xmax; x++) {
	OccElement *el =fog->address(x,y,z);

	// Only consider changing a voxel if it does not 
	// already have a high confidence...
	if (el->totalWeight > 0 && el->totalWeight < 100000) {
	  el->value = MAX(0, MIN(65535, 32767 + 1.2*(el->value - 32767)));
	}
      }
    }
  }

  // set frontRLEGrid from fog, since we made changes in-place
  frontRLEGrid->voxRLE(fog);

  // don't swap fog & bog, because we made changes in-place to fog
  return TCL_OK;
  
}

//////////////////////////////////////////////////////////////////////
// Feb 9, 1999
// Hole filling addition:
//   Remove black/white from volume
//   Get rid of voxels with 65535 or 0 values...
//   works directly on fog, so doesn't need to swapbuffers...

//////////////////////////////////////////////////////////////////////
int
Vrip_RemoveBWVolCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
  if (frontRLEGrid == NULL) {
    interp->result = "Grid not allocated.";
    return TCL_ERROR;
  }

  // Limits so that when we add/subtract 2, we don't walk off the volume
  int xmin=2,ymin=2,zmin=2;
  int xmax=frontRLEGrid->xdim-2;
  int ymax=frontRLEGrid->ydim-2;
  int zmax=frontRLEGrid->zdim-2;

  if (argc>6)
    {
      xmin=MAX(xmin,atoi(argv[1]));
      ymin=MAX(ymin,atoi(argv[2]));
      zmin=MAX(zmin,atoi(argv[3]));
      xmax=MIN(xmax,atoi(argv[4]));
      ymax=MIN(ymax,atoi(argv[5]));
      zmax=MIN(zmax,atoi(argv[6]));
    }
  
  printf ("Removing black & white %d %d, %d %d, %d %d\n",xmin,xmax,ymin,ymax,zmin,zmax);

  // Allocate fog, bog only once, to avoid memory leakage
  if (fog == NULL) fog=frontRLEGrid->voxExpand(); 

  for (int y=ymin;y<ymax;y++) {
    std::cout << y << " ";
    for (int z=zmin;z< zmax; z++) {
      for (int x=xmin;x<xmax; x++) {
	OccElement *el =fog->address(x,y,z);

	// Only consider changing a voxel if it does not 
	// already have a high confidence...
	if (el->totalWeight > 0 && el->totalWeight < 100000) {
	  if (el->value == 0 || el->value == 65535) {
	    // Set to values for "unknown"
	    el->value = 65535;
	    el->totalWeight = 0;
	  }
	}
      }
    }
  }

  // set frontRLEGrid from fog, since we made changes in-place
  frontRLEGrid->voxRLE(fog);

  // don't swap fog & bog, because we made changes in-place to fog
  return TCL_OK;
  
}


int
Vrip_PhotoSliceRLECmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{    
    char *axis;
    OccElement *slice;
    int sliceNum, xdim, ydim;
    float weight, weightScale, value;

    if (argc != 5) {
	interp->result = "Usage: vrip_photoslicerle <photo-widget> <slice-num> <axis> <weight-scale>";
	return TCL_ERROR;
    }    

    if (frontRLEGrid == NULL) {
	interp->result = "Grid not allocated.";
	return TCL_ERROR;
    }
    
    Tk_PhotoHandle handle = Tk_FindPhoto(theInterp, argv[1]);
    if (handle == NULL)
	return TCL_ERROR;

    sliceNum = atoi(argv[2]);
    axis = argv[3];

    weightScale = atof(argv[4]);
    if (weightScale <= 0) {
	interp->result = "weight scale out of range";
	return TCL_ERROR;
    }

    slice = frontRLEGrid->getSlice(axis, sliceNum, &xdim, &ydim);
    
    if (sliceBuf == NULL) {
	sliceBuf = new uchar[3*xdim*ydim];
	sliceXdim = xdim;
	sliceYdim = ydim;
    } 
    else if (xdim*ydim > sliceXdim*sliceYdim) {
	delete sliceBuf;
	sliceBuf = new uchar[3*xdim*ydim];
	sliceXdim = xdim;
	sliceYdim = ydim;
    }

    for (int i = 0; i < xdim*ydim; i++) {
	if (ShowConfSlice) {
	    value = slice[i].value/255.0;
	    if (value >= 255.5) value = 255;
	    sliceBuf[3*i] = uchar(value+0.5);
	    sliceBuf[3*i+1] = uchar(value+0.5);
	    weight = (slice[i].totalWeight & ~OccGridRLE::FALSE_DATA_BIT)
		/weightScale;
	    if (weight > 255.5) weight = 255;
	    sliceBuf[3*i+2] = uchar(weight+0.5);
#if 0
	    value = slice[i].value/255.0;
	    if (value >= 255.5) value = 255;
	    sliceBuf[3*i] = uchar(value+0.5);
	    sliceBuf[3*i+1] = uchar(value+0.5);
	    sliceBuf[3*i+2] = uchar(value+0.5);
#elif 0
	    weight = 8*(slice[i].totalWeight & ~OccGridRLE::FALSE_DATA_BIT)
		/weightScale;
	    if (weight > 255.5) weight = 255;
	    sliceBuf[3*i] = uchar(weight+0.5);
	    sliceBuf[3*i+1] = uchar(weight+0.5);
	    sliceBuf[3*i+2] = 255;
#endif
	}
	else if (ShowValueWeight) {
	    if (slice[i].value == USHRT_MAX && 
		(slice[i].totalWeight & ~OccGridRLE::FALSE_DATA_BIT) == 0) {
#if 1
		sliceBuf[3*i] = 145;
		sliceBuf[3*i+1] = 117;
		sliceBuf[3*i+2] = 74;
#else
		sliceBuf[3*i] = 0;
		sliceBuf[3*i+1] = 0;
		sliceBuf[3*i+2] = 0;
#endif
	    } else {
		weight = (slice[i].totalWeight & ~OccGridRLE::FALSE_DATA_BIT)
		    /weightScale;
		value = slice[i].value/256 - 128;
		value = value*weight/256 + 128;
		if (value >= 255.5)
		   value = 255;
		else if (value < -0.5)
		   value = 0;
		sliceBuf[3*i] = sliceBuf[3*i+1] = sliceBuf[3*i+2] = uchar(value+0.5);
		//sliceBuf[3*i] = sliceBuf[3*i+1] = sliceBuf[3*i+2] = uchar(weight/1.5+0.5);
	    }
	} else {
	    if (slice[i].value == USHRT_MAX && 
		(slice[i].totalWeight & ~OccGridRLE::FALSE_DATA_BIT) == 0) {
#if 1
		sliceBuf[3*i] = 145;
		sliceBuf[3*i+1] = 117;
		sliceBuf[3*i+2] = 74;
#else
		sliceBuf[3*i] = 0;
		sliceBuf[3*i+1] = 0;
		sliceBuf[3*i+2] = 0;
#endif
	    } else {
		value = slice[i].value/255.0;
		if (value >= 255.5) value = 255;
		sliceBuf[3*i] = sliceBuf[3*i+1] = sliceBuf[3*i+2] = uchar(value+0.5);
		//sliceBuf[3*i] = sliceBuf[3*i+1] = sliceBuf[3*i+2] = uchar(weight/1.5+0.5);
	    }
	}
    }

    Tk_PhotoSetSize(handle, xdim, ydim);

    Tk_PhotoImageBlock block;
    
    if (EQSTR(axis, "x")) {
	block.pixelPtr = sliceBuf + 3*xdim*(ydim-1);
	block.pitch = -xdim*3;
    } else if (EQSTR(axis, "y")) {
	block.pixelPtr = sliceBuf;
	block.pitch = xdim*3;
    } else if (EQSTR(axis, "z")) {
	block.pixelPtr = sliceBuf + 3*xdim*(ydim-1);
	block.pitch = -xdim*3;
    }

    block.pixelSize = 3;
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;
    block.width = xdim;
    block.height = ydim;

    Tk_PhotoPutBlock(handle, &block, 0, 0, xdim, ydim);

    return TCL_OK;

}


int
Vrip_PhotoSliceNormRLECmd(ClientData, Tcl_Interp *interp, int argc, 
			 char *argv[])
{    
    char *axis;
    OccNormElement *slice;
    int sliceNum, xdim, ydim;
    float weight, weightScale, value;

    if (argc != 5) {
	interp->result = "Usage: vrip_photoslicerle <photo-widget> <slice-num> <axis> <weight-scale>";
	return TCL_ERROR;
    }    

    if (frontRLEGridNorm == NULL) {
	interp->result = "Grid not allocated.";
	return TCL_ERROR;
    }
    
    Tk_PhotoHandle handle = Tk_FindPhoto(theInterp, argv[1]);
    if (handle == NULL)
	return TCL_ERROR;

    sliceNum = atoi(argv[2]);
    axis = argv[3];

    weightScale = atof(argv[4]);
    if (weightScale <= 0) {
	interp->result = "weight scale out of range";
	return TCL_ERROR;
    }

    slice = frontRLEGridNorm->getSlice(axis, sliceNum, &xdim, &ydim);
    
    if (sliceBuf == NULL) {
	sliceBuf = new uchar[3*xdim*ydim];
	sliceXdim = xdim;
	sliceYdim = ydim;
    } 
    else if (xdim*ydim > sliceXdim*sliceYdim) {
	delete sliceBuf;
	sliceBuf = new uchar[3*xdim*ydim];
	sliceXdim = xdim;
	sliceYdim = ydim;
    }

    if (!ShowNormals) {
	for (int i = 0; i < xdim*ydim; i++) {
	    value = slice[i].value/255.0;
	    if (value >= 255.5) value = 255;
	    sliceBuf[3*i] = uchar(value+0.5);
	    weight = slice[i].totalWeight/weightScale;
	    if (weight > 255.5) weight = 255;
	    sliceBuf[3*i+1] = uchar(weight+0.5);
	}
    } else {
	for (int i = 0; i < xdim*ydim; i++) {
	    sliceBuf[3*i] = uchar(int(slice[i].nx)+127);
	    sliceBuf[3*i+1] = uchar(int(slice[i].ny)+127);
	    sliceBuf[3*i+2] = uchar(int(slice[i].nz)+127);
	}
    }

    Tk_PhotoSetSize(handle, xdim, ydim);

    Tk_PhotoImageBlock block;
    
    if (EQSTR(axis, "x")) {
	block.pixelPtr = sliceBuf + 3*xdim*(ydim-1);
	block.pitch = -xdim*3;
    } else if (EQSTR(axis, "y")) {
	block.pixelPtr = sliceBuf;
	block.pitch = xdim*3;
    } else if (EQSTR(axis, "z")) {
	block.pixelPtr = sliceBuf + 3*xdim*(ydim-1);
	block.pitch = -xdim*3;
    }

    block.pixelSize = 3;

    if (!ShowNormals) {
	block.offset[0] = 0;
	block.offset[1] = 0;
	block.offset[2] = 1;
    } else {
	block.offset[0] = 0;
	block.offset[1] = 1;
	block.offset[2] = 2;
    }

    block.width = xdim;
    block.height = ydim;

    Tk_PhotoPutBlock(handle, &block, 0, 0, xdim, ydim);

    return TCL_OK;

}



int
Vrip_GetVoxelRLECmd(ClientData, Tcl_Interp *interp, int argc, char *argv[]) 
{
    int ix, iy, iz;
    char buf[512];

    if (argc != 4) {
        interp->result = "Usage: vrip_getvoxelrle <ix> <iy> <iz>";
        return TCL_ERROR;
    }

    if (frontRLEGrid == NULL) {
	interp->result = "Grid not allocated.";
	return TCL_ERROR;
    }

    ix = atoi(argv[1]);
    iy = atoi(argv[2]);
    iz = atoi(argv[3]);

    if (ix >= 0 && ix < frontRLEGrid->xdim &&
	iy >= 0 && iy < frontRLEGrid->ydim &&
	iz >= 0 && iz < frontRLEGrid->zdim) {
       OccElement *el = frontRLEGrid->getElement(ix, iy, iz);
       sprintf(buf, "(%.3d, %.3d, %.3d): v = %+.5d, wt = %.5d",
	       ix, iy, iz, (int) el->value + SHRT_MIN, el->totalWeight);
       }
    else
       sprintf(buf, "<index out of range>");

    Tcl_SetResult(interp, buf, TCL_VOLATILE);

    return TCL_OK;
}



int
Vrip_SetRenderPhotoCmd(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (theRenderPhoto != NULL)
	delete [] theRenderPhoto;

    if (argc == 1) {
	theRenderPhoto = NULL;
	return TCL_OK;
    }
    else {
	theRenderPhoto = new char[strlen(argv[1])+1];
	strcpy(theRenderPhoto, argv[1]);
	return TCL_OK;
    }
}



int
updateRenderPhoto(DepthMap *dm)
{
    int xdim, ydim;

    if (theRenderPhoto == NULL)
	return 0;

    Tk_PhotoHandle handle = Tk_FindPhoto(theInterp, theRenderPhoto);
    if (handle == NULL)
	return 0;

    xdim = dm->xdim;
    ydim = dm->ydim;

    if (renderBuf == NULL) {
	renderBuf = new uchar[xdim*ydim];
	renderXdim = xdim;
	renderYdim = ydim;
    } 
    else if (xdim*ydim > renderXdim*renderYdim) {
	delete renderBuf;
	renderBuf = new uchar[xdim*ydim];
	renderXdim = xdim;
	renderYdim = ydim;
    }

    for (int i = 0; i < xdim*ydim; i++) {
	renderBuf[i] = uchar(fabs(dm->elems[i].conf*255+0.5));
    }

    Tk_PhotoSetSize(handle, xdim, ydim);

    Tk_PhotoImageBlock block;
    
    block.pixelPtr = renderBuf + xdim*(ydim-1);
    block.pitch = -xdim;
    block.pixelSize = 1;
    block.offset[0] = 0;
    block.offset[1] = 0;
    block.offset[2] = 0;
    block.width = xdim;
    block.height = ydim;

    Tk_PhotoPutBlock(handle, &block, 0, 0, xdim, ydim);

    return 1;

}


